(* ========================================================================= *)
(*                                                                           *)
(*                        COMPOSED_SYSTEMS                                   *)
(*                        QUASI-OPTICAL SYSTEMS                              *)
(*        (c) Copyright, Umair Siddique  (2012-2015)                         *)
(*                       Hardware Verification Group,                        *)
(*                       Concordia University                                *)
(*                                                                           *)
(*     Contact: <muh_sidd@ece.concordia.ca>                                  *)
(*                                                                           *)
(* Last update: Dec 2015                                                     *)
(*                                                                           *)
(* ========================================================================= *)

needs "q_parameters_beams.ml";;


new_type_abbrev("composed_system",`:optical_system list`);;

let composed_system = define 
   `composed_system ([]:optical_system list)  = 
              cmat2x2 (Cx(&1)) (Cx(&0))  (Cx(&0)) (Cx(&1)) /\
    composed_system ((CONS sys) cs) = 
    composed_system cs ** system_composition sys `;;


let is_valid_composed_system = new_definition
   `is_valid_composed_system  (sys:optical_system list) <=>
    ALL is_valid_optical_system sys `;;

new_type_abbrev("gbeam",`:beam list`);;


let fst_sing_beam = new_definition
  `fst_sing_beam ((sr1,sr2,rs):ray )= sr1`;; 


let last_sing_beam = define `
last_sing_beam ((br1,br2,[]):beam) = br2 /\
     last_sing_beam (br1,br2,CONS bs3 t) = SND (LAST (CONS bs3 t))`;;

let LAST_SINGLE_BEAM_REC = prove
  (`!sr1 sr2 sr3 sr4 t.
    last_sing_beam (sr1,sr2,[]) = sr2 /\
    last_sing_beam (sr1,sr2,CONS (sr3,sr4) t) = last_sing_beam (sr3,sr4,t)`,
  REPEAT (LIST_INDUCT_TAC ORELSE GEN_TAC)
  THEN SIMP_TAC[last_sing_beam;LAST;NOT_CONS_NIL]);;

(*
let is_valid_genbeam = define `
    (is_valid_genbeam ([]:gbeam) <=> F) /\ 
     (is_valid_genbeam (CONS h t) <=> 
     ((last_sing_beam h = fst_sing_beam (HD t)) /\ 
      is_valid_genbeam t ))`;;
*)

let is_valid_gen_beam = define `
    (is_valid_gen_beam ([]:gbeam) <=> F) /\ 
     (is_valid_gen_beam (CONS h t) <=> 
     ((last_single_ray h = fst_sing_beam (HD t)) /\ 
      is_valid_gen_beam t ))`;;


let is_valid_gbeam_in_c_system = define `
   (is_valid_gbeam_in_c_system (beam:gbeam) lam ([]:optical_system list) 
    <=> F) /\ 
   (is_valid_gbeam_in_c_system  [] lam (sys:optical_system list) 
    <=> F) /\ 
   (is_valid_gbeam_in_c_system (CONS r1 rs) lam (CONS h hs) 
    <=> ((is_valid_beam_in_system r1 lam h) /\ 
       ~((((system_composition h)$2$1):complex)*
        qq(FST(fst_sing_beam r1),SND(fst_sing_beam r1),lam) + 
        (((system_composition h)$2$2):complex) = Cx(&0) )/\ 
          (is_valid_gbeam_in_c_system rs lam hs)))`;;


let sys_constraints = new_definition `
    sys_constraints (sys:composed_system) (beam:gbeam) (lam:real) =
    (&0 < lam /\
         is_valid_composed_system sys /\ 
         is_valid_gen_beam (beam)/\
         is_valid_gbeam_in_c_system beam lam sys)`;;



let ABCD_COMPOSED_SYSTEM = prove
 (`!(sys:composed_system) (gbeam:gbeam)  aa bb cc dd.
         &0 < lam /\
         is_valid_composed_system sys /\ 
         is_valid_gen_beam (gbeam)/\
         is_valid_gbeam_in_c_system gbeam lam sys  /\ 
         cmat2x2 aa bb cc dd = composed_system sys
         ==> (let (z,w0) = fst_sing_beam (HD (gbeam)) and
                 (z_n,w0_n) = last_single_ray (LAST (gbeam)) in

              qq (z_n,w0_n,lam) =
              (aa * qq (z,w0,lam) + bb) / (cc * qq (z,w0,lam) + dd))`,
  LIST_INDUCT_TAC THENL
   [REWRITE_TAC
      [is_valid_gbeam_in_c_system];
    ALL_TAC] THEN
  REWRITE_TAC
    [ALL;is_valid_composed_system;
     GSYM CMATRIX_VECTOR_MUL_ASSOC;composed_system] THEN
  REPEAT (LIST_INDUCT_TAC ORELSE GEN_TAC) THEN
  REWRITE_TAC
    [is_valid_gbeam_in_c_system] THEN
  REWRITE_TAC
    [HD;fst_sing_beam;last_sing_beam;
     ALL; is_valid_optical_system;is_valid_gen_beam] THEN
  REPEAT STRIP_TAC THEN REPEAT LET_TAC THEN
  UNDISCH_TAC
    `!gbeam aa bb cc dd.
          &0 < lam /\
          is_valid_composed_system t /\
          is_valid_gen_beam gbeam /\
          is_valid_gbeam_in_c_system gbeam lam t /\
          cmat2x2 aa bb cc dd = composed_system t
          ==> (let z,w0 = fst_sing_beam (HD gbeam)
               and z_n,w0_n = last_single_ray (LAST gbeam) in
               qq (z_n,w0_n,lam) =
               (aa * qq (z,w0,lam) + bb) / (cc * qq (z,w0,lam) + dd))` THEN
  DISCH_THEN (MP_TAC o SPEC `t':gbeam`) THEN
  DISCH_THEN (MP_TAC o SPEC `((composed_system t)$1$1):complex`) THEN
  DISCH_THEN (MP_TAC o SPEC `((composed_system t)$1$2):complex`) THEN
  DISCH_THEN (MP_TAC o SPEC `((composed_system t)$2$1):complex`) THEN
  DISCH_THEN (MP_TAC o SPEC `((composed_system t)$2$2):complex`) THEN
  ANTS_TAC THENL
   [ASM_SIMP_TAC
      [is_valid_composed_system] THEN
    CCOMMON_TAC [];
    ALL_TAC] THEN
  REPEAT LET_TAC THEN MP_TAC ABCD_LAW_OPTICAL_SYSTEM THEN
  DISCH_THEN (MP_TAC o SPEC `h:optical_system`) THEN
  DISCH_THEN (MP_TAC o SPEC `h':beam`) THEN
  DISCH_THEN (MP_TAC o SPEC `((system_composition h)$1$1):complex`) THEN
  DISCH_THEN (MP_TAC o SPEC `((system_composition h)$1$2) :complex`) THEN
  DISCH_THEN (MP_TAC o SPEC `((system_composition h)$2$1):complex`) THEN
  DISCH_THEN (MP_TAC o SPEC `((system_composition h)$2$2) :complex`) THEN
  ANTS_TAC THENL [ASM_SIMP_TAC [] THEN CCOMMON_TAC []; ALL_TAC] THEN
  REPEAT LET_TAC THEN
  UNDISCH_TAC `fst_sing_beam ((z'',w0''),(z''',w0'''),rs) = z,w0` THEN
  REWRITE_TAC [fst_sing_beam] THEN
  UNDISCH_TAC `(z_n'':real),(w0_n'':real) = (z':real),(w0':real)` THEN
  SIMP_TAC [PAIR_EQ] THEN DISCH_TAC THEN SIMP_TAC [EQ_SYM_EQ] THEN
  ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN
  ASSERT_TAC
    `last_single_ray (LAST (CONS ((z'',w0''),(z''',w0'''),rs) t')) = 
last_single_ray (LAST t')` THENL
   [REPEAT (POP_ASSUM MP_TAC) THEN
    ASM_CASES_TAC
      `t' :((real#real)#(real#real)#((real#real)#real#real)list)list = 
               []` THENL
     [ONCE_ASM_REWRITE_TAC [] THEN
      SIMP_TAC
        [is_valid_gen_beam];
      ASM_SIMP_TAC [LAST]];
    ALL_TAC] THEN
  POP_ASSUM MP_TAC THEN ASM_REWRITE_TAC [] THEN DISCH_TAC THEN DISCH_TAC THEN
  DISCH_TAC THEN DISCH_TAC THEN
  UNDISCH_TAC `(z_n:real),(w0_n:real) = (z_n':real),(w0_n':real)` THEN
  SIMP_TAC [PAIR_EQ] THEN POP_ASSUM MP_TAC THEN SIMP_TAC [] THEN
  REPEAT STRIP_TAC THEN MATCH_MP_TAC Q_ABCD_ITERATE_2 THEN
  EXISTS_TAC `((system_composition h)$1$1):complex` THEN
  EXISTS_TAC `((system_composition h)$1$2) :complex` THEN
  EXISTS_TAC `((system_composition h)$2$1):complex` THEN
  EXISTS_TAC `((system_composition h)$2$2) :complex` THEN ASM_SIMP_TAC [] THEN
  SIMP_TAC [GSYM CMAT_2X2_EQ_COMPONENT] THEN
  UNDISCH_TAC
    `~(system_composition h$2$1 * qq (FST (z,w0),SND (z,w0),lam) +
        system_composition h$2$2 =
        Cx (&0))` THEN
  SIMP_TAC [FST; SND]);;



let beam_origin = new_definition `
    beam_origin (beam:gbeam) = fst_sing_beam (HD beam)`;;

let beam_end = new_definition `
    beam_end (beam:gbeam) = last_single_ray (LAST beam)`;;


let ABCD_COMPOSED_SYSTEM_ALT = prove (`!sys gbeam lam aa bb cc dd.
         &0 < lam /\
         is_valid_composed_system sys /\
         is_valid_gen_beam gbeam /\
         is_valid_gbeam_in_c_system gbeam lam sys /\
         cmat2x2 aa bb cc dd = composed_system sys
         ==> (let z,w0 = beam_origin gbeam 
              and z_n,w0_n = beam_end gbeam in
              qq (z_n,w0_n,lam) =
              (aa * qq (z,w0,lam) + bb) / (cc * qq (z,w0,lam) + dd))`,

MESON_TAC[ ABCD_COMPOSED_SYSTEM; beam_origin;beam_end]);;



let ABCD_COMPOSED_SYSTEM_ALT_2 = prove (`!sys gbeam lam A B C D.
         sys_constraints sys gbeam lam /\
         cmat2x2 A B C D = composed_system sys
         ==> (let z,w0 = beam_origin gbeam 
              and z_n,w0_n = beam_end gbeam in
              qq (z_n,w0_n,lam) =
              (A * qq (z,w0,lam) + B) / (C * qq (z,w0,lam) + D))`,

REPEAT STRIP_TAC THEN 
MATCH_MP_TAC ABCD_COMPOSED_SYSTEM_ALT THEN 
EXISTS_TAC ` sys:composed_system` THEN 
REPEAT(POP_ASSUM MP_TAC) THEN 
SIMP_TAC[sys_constraints]);;



let Q_REARRANGEMENT = 
    COMPLEX_FIELD`   ~((C * q1 + D) = Cx(&0))   /\ 
     ~((A * q1 + B) = Cx(&0))   /\
     ~(q2 = Cx(&0)) /\ ~(q1 = Cx(&0))==> 
      (q2 = (A * q1 + B) / (C * q1 + D) <=>
           inv q2 = (C * q1 + D)/ (A * q1 + B))`;;

let Q_REARRANGEMENT_2 = 
    COMPLEX_FIELD` ! q1 q2 A B C D.  ~((C * q1 + D) = Cx(&0))   /\ 
     ~((A * q1 + B) = Cx(&0))   /\
     ~(q2 = Cx(&0)) /\ ~(q1 = Cx(&0))==> 
      (q2 = (A * q1 + B) / (C * q1 + D) <=>
           inv q2 = (C  + D / q1)/ (A  + B / q1))`;;


let ABCD_COMPOSED_SYSTEM_ALTERNATIVE_FORM = prove 
(`!sys gbeam lam A B C D.
         sys_constraints sys gbeam lam /\
         cmat2x2 A B C D = composed_system sys /\
         &0 < SND (beam_origin gbeam) /\
         &0 < SND (beam_end gbeam)  /\
       ~(A * qq (FST(beam_origin gbeam),SND (beam_origin gbeam),lam)+B = Cx (&0))/\
       ~(C * qq (FST(beam_origin gbeam),SND (beam_origin gbeam),lam)+D = Cx (&0))
       ==> (let z,w0 = beam_origin gbeam 
              and z_n,w0_n = beam_end gbeam in
              inv (qq (z_n,w0_n,lam)) =
            (C  + D / qq (z,w0,lam)) / (A  + B /  qq (z,w0,lam))  )`,

REPEAT STRIP_TAC THEN 
REPEAT LET_TAC THEN 
IMP_REWRITE_TAC[GSYM Q_REARRANGEMENT_2] THEN 
MP_TAC (ISPECL[`sys:composed_system`; `gbeam:gbeam`; `lam:real`; 
                 `A:complex`; `B:complex`; `C:complex` ;`D:complex`] 
                ABCD_COMPOSED_SYSTEM_ALT_2) THEN
ASM_SIMP_TAC[] THEN
REPEAT LET_TAC THEN
SIMP_TAC[] THEN
REPEAT(POP_ASSUM MP_TAC) THEN
REWRITE_TAC[sys_constraints] THEN
SIMP_TAC[Q_NZ;SND;FST;sys_constraints] THEN
SIMP_TAC[EQ_SYM_EQ;sys_constraints;Q_NZ]);;



