(* ========================================================================= *)
(*                                                                           *)
(*                       		                                     *)
(*                        QUASI-OPTICAL SYSTEMS                              *)
(*        (c) Copyright, Umair Siddique  (2012-2015)                         *)
(*                       Hardware Verification Group,                        *)
(*                       Concordia University                                *)
(*                                                                           *)
(*     Contact: <muh_sidd@ece.concordia.ca>                                  *)
(*                                                                           *)
(* Last update: Dec 2015                                                     *)
(*                                                                           *)
(* ========================================================================= *)


needs "composed_system_gaussian.ml";;


(* ========================================================================= *)
(*                 General Quasi-Optical System                              *)
(* ========================================================================= *)

let asm_real_prove = REPEAT(POP_ASSUM MP_TAC) THEN CONV_TAC REAL_FIELD;;


let quasi_optical_system = new_definition` 
   quasi_optical_system (sys:optical_system) din dout ni no = 
    [([],ni,din);sys;([],no,dout)]`;;

let QUASI_OPTICAL_SYSTEM_MAT = prove(` 
    !sys din dout ni no A B C D. 
    ((system_composition sys) = cmat2x2 A B C D) ==> 
     composed_system (quasi_optical_system sys din dout ni no) = 
     cmat2x2 (A + C*(Cx(dout))) (A*Cx(din) + B + C*Cx(din)*Cx(dout) + D*Cx(dout)) 
            (C) (C*Cx(din) + D)`, 

REPEAT STRIP_TAC THEN
REWRITE_TAC[composed_system;quasi_optical_system;
            system_composition;CMAT2X2_LID;free_space_matrix] THEN
ASM_REWRITE_TAC[] THEN SIMP_TAC[ CMAT2X2_MUL;CMAT2X2_EQ] THEN
CONV_TAC COMPLEX_FIELD);;


let QUASI_OPTICAL_SYSTEM_ABCD_LAW = prove (
   `!sys gbeam lam A B C D.
   &0 < lam /\
   is_valid_composed_system (quasi_optical_system sys din dout ni no)/\
   is_valid_gen_beam gbeam /\
   is_valid_gbeam_in_c_system gbeam lam (quasi_optical_system sys din dout ni no)
   /\ cmat2x2 A B C D = (system_composition sys)
         ==> (let z,w0_in = beam_origin gbeam and 
                  z_n,w0_end = beam_end gbeam in
        qq (z_n,w0_end,lam) =
       ((A + C * Cx dout) * qq (z,w0_in,lam) + 
       (A * Cx din + B + C * Cx din * Cx dout + D * Cx dout)) / 
       (C * qq (z,w0_in,lam) + (C * Cx din + D)))`,

REPEAT STRIP_TAC THEN 
MATCH_MP_TAC ABCD_COMPOSED_SYSTEM_ALT THEN 
EXISTS_TAC `(quasi_optical_system sys din dout ni no)` THEN 
ASM_SIMP_TAC[] THEN 
IMP_REWRITE_TAC[QUASI_OPTICAL_SYSTEM_MAT] THEN 
EXISTS_TAC `A:complex` THEN 
EXISTS_TAC `B:complex` THEN 
EXISTS_TAC `C:complex` THEN 
EXISTS_TAC `D:complex` THEN 
ASM_MESON_TAC[EQ_SYM_EQ]);;



let QUASI_OPTICAL_SYSTEM_ABCD_LAW_ALT = prove (
   `!sys gbeam lam A B C D.
   sys_constraints (quasi_optical_system sys din dout ni no) gbeam lam /\ 
   cmat2x2 A B C D = (system_composition sys)
         ==> (let z,w0_in = beam_origin gbeam and 
                  z_n,w0_end = beam_end gbeam in
        qq (z_n,w0_end,lam) =
       ((A + C * Cx dout) * qq (z,w0_in,lam) + 
       (A * Cx din + B + C * Cx din * Cx dout + D * Cx dout)) / 
       (C * qq (z,w0_in,lam) + (C * Cx din + D)))`,

REPEAT STRIP_TAC THEN 
MATCH_MP_TAC ABCD_COMPOSED_SYSTEM_ALT_2 THEN 
EXISTS_TAC `(quasi_optical_system sys din dout ni no)` THEN 
ASM_SIMP_TAC[] THEN 
ASM_SIMP_TAC[QUASI_OPTICAL_SYSTEM_MAT ]);;



let Q_REAL_ZERO = prove (`!z w0 lam.
    Re  (qq (z,w0,lam)) = &0 ==>
         qq (z,w0,lam) =  ii * Cx (rayleigh_range w0 lam)`,

SIMP_TAC[qq;GSYM COMPLEX_TRAD] THEN 
SIMP_TAC[RE;COMPLEX_ADD_LID; COMPLEX_TRAD ]);;


let QUASI_OPTICAL_SYSTEM_BEAM_TRANSFORMATION = prove
 (`!sys gbeam lam A B C D.
         &0 < lam /\
         is_valid_composed_system (quasi_optical_system sys din dout ni no) /\
         is_valid_gen_beam gbeam /\
         is_valid_gbeam_in_c_system gbeam lam
         (quasi_optical_system sys din dout ni no) /\
         cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D) = system_composition sys
          ==> 
         (let z,w0_in = beam_origin gbeam and 
         z_n,w0_out = beam_end gbeam in  
         (Re  (qq (z,w0_in,lam)) = (&0)) /\ 
         (Re  (qq (z_n,w0_out,lam)) = (&0)) /\
          ~((C * din + D) pow 2 + (C * rayleigh_range w0_in lam) pow 2 = &0)
          ==>
(dout = 
--((A * din + B) * (C * din + D) + A * C * rayleigh_range w0_in lam pow 2) /
   ((C * din + D) pow 2 + (C * rayleigh_range w0_in lam) pow 2) /\

w0_out pow 2 =
 ((A * D - B * C) * w0_in pow 2) /
 ((C * din + D) pow 2 + (C * w0_in pow 2 * pi / lam) pow 2)))`,
  
  REPEAT GEN_TAC THEN REPEAT DISCH_TAC THEN
  MP_TAC
    (ISPECL
       [`sys:optical_system`; `gbeam:gbeam`; `lam:real`; `(Cx A)`; `(Cx B)`;
        `(Cx C)`; `(Cx D)`]
       QUASI_OPTICAL_SYSTEM_ABCD_LAW) THEN
  ASM_SIMP_TAC [] THEN REPEAT LET_TAC THEN REPEAT DISCH_TAC THEN
  UNDISCH_TAC
    `qq (z_n',w0_end,lam) =
      ((Cx A + Cx C * Cx dout) * qq (z,w0_in,lam) +
       Cx A * Cx din +
       Cx B +
       Cx C * Cx din * Cx dout +
       Cx D * Cx dout) /
      (Cx C * qq (z,w0_in,lam) + Cx C * Cx din + Cx D)` THEN
  UNDISCH_TAC `(z_n:real),(w0_out:real) = z_n',w0_end` THEN
  ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN SIMP_TAC [PAIR_EQ] THEN DISCH_TAC THEN
  ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN ASM_SIMP_TAC [Q_REAL_ZERO] THEN
  REWRITE_TAC [complex_div] THEN REWRITE_TAC [COMPLEX_INV_CNJ] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `(Cx C * ii * Cx (rayleigh_range w0 lam) + Cx C * Cx din + Cx D) =
              Cx C * Cx din + Cx D  +  ii * Cx C * Cx (rayleigh_range w0 lam)`] THEN
  REWRITE_TAC [GSYM CX_MUL; GSYM CX_ADD] THEN REWRITE_TAC [COMPLEX_ADD_ASSOC] THEN
  REWRITE_TAC [GSYM CX_MUL; GSYM CX_ADD] THEN SIMP_TAC [CNJ_ADD; CNJ_CX] THEN
  SIMP_TAC [CNJ_MUL; CNJ_CX; CNJ_II] THEN SIMP_TAC [complex_norm] THEN
  SIMP_TAC [GSYM COMPLEX_TRAD; RE; IM] THEN
  IMP_REWRITE_TAC
    [REAL_FIELD `&0 <= a  /\ &0 <= b ==> &0 <=  a+ b`; REAL_LE_POW_2; CX_SQRT] THEN
  SIMP_TAC [CSQRT] THEN SIMP_TAC [COMPLEX_FIELD `a * b / c = (a*b) / c`] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `(Cx (A + C * dout) * ii * Cx (rayleigh_range w0 lam) +
   Cx (A * din + B + C * din * dout + D * dout)) = 

Cx (A * din + B + C * din * dout + D * dout) + 
ii* Cx (A + C * dout)  * Cx (rayleigh_range w0 lam)`] THEN
  REWRITE_TAC [GSYM CX_MUL] THEN
  SIMP_TAC [COMPLEX_FIELD `(a + b)*(c + d) = a*c + b*c + a*d + b*d`] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `(ii * Cx ((A + C * dout) * rayleigh_range w0 lam)) *
  --ii *
  Cx (C * rayleigh_range w0 lam) = (--(ii pow 2) * Cx ((A + C * dout) * rayleigh_range w0 lam)) *
  Cx (C * rayleigh_range w0 lam)`] THEN
  SIMP_TAC [COMPLEX_POW_II_2] THEN SIMP_TAC [COMPLEX_FIELD `-- --a*b = a*b`] THEN
  SIMP_TAC [COMPLEX_MUL_LID] THEN REWRITE_TAC [GSYM CX_MUL] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `(Cx ((A * din + B + C * din * dout + D * dout) * (C * din + D)) +
  (ii * Cx ((A + C * dout) * rayleigh_range w0 lam)) * Cx (C * din + D) +
  Cx (A * din + B + C * din * dout + D * dout) *
  --ii *
  Cx (C * rayleigh_range w0 lam) +
  Cx (((A + C * dout) * rayleigh_range w0 lam) * C * rayleigh_range w0 lam)) =

Cx ((A * din + B + C * din * dout + D * dout) * (C * din + D)) + 
Cx (((A + C * dout) * rayleigh_range w0 lam) * C * rayleigh_range w0 lam) + 
 ii*((Cx ((A + C * dout) * rayleigh_range w0 lam)) * Cx (C * din + D) -
 Cx (A * din + B + C * din * dout + D * dout) *
  Cx (C * rayleigh_range w0 lam))`] THEN
  REWRITE_TAC [GSYM CX_MUL; GSYM CX_ADD; GSYM CX_SUB] THEN
  REWRITE_TAC [COMPLEX_ADD_ASSOC] THEN REWRITE_TAC [GSYM CX_ADD] THEN
  SIMP_TAC
    [REAL_FIELD
       `((A * din + B + C * din * dout + D * dout) * (C * din + D) +
   ((A + C * dout) * rayleigh_range w0 lam) * C * rayleigh_range w0 lam) =

(A*din + B)*(C*din + D) + A*C  *  (rayleigh_range w0 lam) pow 2
+ dout *((C * din + D) pow 2 + (C * rayleigh_range w0 lam) pow 2)`] THEN
  SIMP_TAC
    [REAL_FIELD
       `(((A + C * dout) * rayleigh_range w0 lam) * (C * din + D) -
      (A * din + B + C * din * dout + D * dout) * C * rayleigh_range w0 lam) = 
(A * D  - B*C) * rayleigh_range w0 lam`] THEN
  SIMP_TAC [COMPLEX_FIELD `(a + ii*b) / c = a / c + ii * (b / c)`] THEN
  REWRITE_TAC [GSYM CX_DIV; GSYM COMPLEX_TRAD] THEN
  SIMP_TAC [COMPLEX_EQ; RE; IM] THEN
  SIMP_TAC [IM_MUL_II; RE_MUL_II; IM_CX; RE_CX] THEN
  SIMP_TAC
    [REAL_FIELD
       `-- &0 =
 ((A * din + B) * (C * din + D) +
  A * C * rayleigh_range w0_in lam pow 2 +
  dout * ((C * din + D) pow 2 + (C * rayleigh_range w0_in lam) pow 2)) /
 ((C * din + D) pow 2 + (C * rayleigh_range w0_in lam) pow 2) <=> 

((A * din + B) * (C * din + D) +
  A * C * rayleigh_range w0_in lam pow 2) /
((C * din + D) pow 2 + (C * rayleigh_range w0_in lam) pow 2) +

dout * ((C * din + D) pow 2 + (C * rayleigh_range w0_in lam) pow 2) /
       ((C * din + D) pow 2 + (C * rayleigh_range w0_in lam) pow 2) = &0`] THEN
  IMP_REWRITE_TAC [REAL_DIV_REFL] THEN ASM_SIMP_TAC [REAL_MUL_RID] THEN
  SIMP_TAC
    [REAL_FIELD
       `((A * din + B) * (C * din + D) + A * C * rayleigh_range w0_in lam pow 2) /
((C * din + D) pow 2 + (C * rayleigh_range w0_in lam) pow 2) + dout = &0 <=>
dout =
     --((A * din + B) * (C * din + D) +
        A * C * rayleigh_range w0_in lam pow 2) /
     ((C * din + D) pow 2 + (C * rayleigh_range w0_in lam) pow 2)`] THEN
  SIMP_TAC [MESON [] `a /\ b ==> c <=> a ==> b ==> c`] THEN DISCH_TAC THEN
  REWRITE_TAC [rayleigh_range] THEN
  SIMP_TAC
    [REAL_FIELD `(pi * w0_out pow 2) / lam = (w0_out pow 2)* pi / lam`] THEN
  IMP_REWRITE_TAC
    [REAL_FIELD
       `!(a:real) b c d.  ~(b = &0) ==> 
    ((a * b) = (c * e * b) / d) =   ((a) = (c * e) / d)`] THEN
  IMP_REWRITE_TAC
    [REAL_FIELD `!(a:real) b c d.  &0 < a /\ &0 < b  ==> ~(a / b = &0)`;
     PI_POS] THEN
  ASM_SIMP_TAC []);;


let QUASI_OPTICAL_SYSTEM_BEAM_TRANSFORMATION_ALT = prove(
    ` !sys gbeam din dout lam  ni no w0_in w0_out z z_n A B C D. 
    (sys_constraints (quasi_optical_system sys din dout ni no) gbeam lam) /\
    (cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D) = system_composition sys) /\
    (z,w0_in = beam_origin gbeam) /\
    (z_n,w0_out = beam_end gbeam) /\
    (Re (qq (z,w0_in,lam)) = &0) /\
    (Re (qq (z_n,w0_out,lam)) = &0) /\
    ~((C * din + D) pow 2 + (C * rayleigh_range w0_in lam) pow 2 = &0)
     ==>
dout =
 --((A * din + B) * (C * din + D) + A * C * rayleigh_range w0_in lam pow 2) /
   ((C * din + D) pow 2 + (C * rayleigh_range w0_in lam) pow 2) /\

w0_out pow 2 =
 ((A * D - B * C) * w0_in pow 2) /
 ((C * din + D) pow 2 + (C * w0_in pow 2 * pi / lam) pow 2)`,

REWRITE_TAC[ sys_constraints] THEN 
REPEAT STRIP_TAC THENL[
MP_TAC (ISPECL[`sys:optical_system`;`gbeam:gbeam`;`lam:real`;`A:real`;
  `B:real`;`C:real`;`D:real`]QUASI_OPTICAL_SYSTEM_BEAM_TRANSFORMATION) THEN 
ASM_SIMP_TAC[] THEN 
REPEAT LET_TAC THEN 
REPEAT(POP_ASSUM MP_TAC) THEN
SIMP_TAC[PAIR_EQ]; ALL_TAC] THEN 
MP_TAC (ISPECL[`sys:optical_system`;`gbeam:gbeam`;`lam:real`;`A:real`;
  `B:real`;`C:real`;`D:real`]QUASI_OPTICAL_SYSTEM_BEAM_TRANSFORMATION) THEN 
ASM_SIMP_TAC[] THEN 
REPEAT LET_TAC THEN 
REPEAT(POP_ASSUM MP_TAC) THEN
SIMP_TAC[PAIR_EQ]);;


let QUASI_OPTICAL_SYSTEM_BEAM_TRANSFORMATION_DOUT = prove(
    ` !sys gbeam din dout lam  ni no w0_in w0_out z z_n A B C D. 
    (sys_constraints (quasi_optical_system sys din dout ni no) gbeam lam) /\
    (cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D) = system_composition sys) /\
    (z,w0_in = beam_origin gbeam) /\
    (z_n,w0_out = beam_end gbeam) /\
    (Re (qq (z,w0_in,lam)) = &0) /\
    (Re (qq (z_n,w0_out,lam)) = &0) /\
    ~((C * din + D) pow 2 + (C * rayleigh_range w0_in lam) pow 2 = &0)
     ==>
dout =
 --((A * din + B) * (C * din + D) + A * C * rayleigh_range w0_in lam pow 2) /
   ((C * din + D) pow 2 + (C * rayleigh_range w0_in lam) pow 2)`,

REWRITE_TAC[ sys_constraints] THEN 
REPEAT STRIP_TAC THEN
MP_TAC (ISPECL[`sys:optical_system`;`gbeam:gbeam`;`lam:real`;`A:real`;
  `B:real`;`C:real`;`D:real`]QUASI_OPTICAL_SYSTEM_BEAM_TRANSFORMATION) THEN 
ASM_SIMP_TAC[] THEN 
REPEAT LET_TAC THEN 
REPEAT(POP_ASSUM MP_TAC) THEN
SIMP_TAC[PAIR_EQ]);;

let QUASI_OPTICAL_SYSTEM_BEAM_TRANSFORMATION_WOUT = prove(
    ` !sys gbeam din dout lam  ni no w0_in w0_out z z_n A B C D. 
    (sys_constraints (quasi_optical_system sys din dout ni no) gbeam lam) /\
    (cmat2x2 (Cx A) (Cx B) (Cx C) (Cx D) = system_composition sys) /\
    (z,w0_in = beam_origin gbeam) /\
    (z_n,w0_out = beam_end gbeam) /\
    (Re (qq (z,w0_in,lam)) = &0) /\
    (Re (qq (z_n,w0_out,lam)) = &0) /\
    ~((C * din + D) pow 2 + (C * rayleigh_range w0_in lam) pow 2 = &0)
     ==>
w0_out pow 2 =
 ((A * D - B * C) * w0_in pow 2) /
 ((C * din + D) pow 2 + (C * w0_in pow 2 * pi / lam) pow 2)`,

REWRITE_TAC[ sys_constraints] THEN 
REPEAT STRIP_TAC THEN
MP_TAC (ISPECL[`sys:optical_system`;`gbeam:gbeam`;`lam:real`;`A:real`;
  `B:real`;`C:real`;`D:real`]QUASI_OPTICAL_SYSTEM_BEAM_TRANSFORMATION) THEN 
ASM_SIMP_TAC[] THEN 
REPEAT LET_TAC THEN 
REPEAT(POP_ASSUM MP_TAC) THEN
SIMP_TAC[PAIR_EQ]);;


let QUASI_OPTICAL_SYSTEM_BEAM_TRANSFORMATION_DOUT_WOUT = 
 CONJ QUASI_OPTICAL_SYSTEM_BEAM_TRANSFORMATION_DOUT
      QUASI_OPTICAL_SYSTEM_BEAM_TRANSFORMATION_WOUT;;


(*=======================================================*)
(*==============  APEX TELESCOPE ========================*)
(*=======================================================*)


let thin_lens = new_definition
  `thin_lens R1 R2 n1 n2: optical_system =
    ([(n1,&0),(spherical R1),transmitted;
      (n2,&0),(spherical R2),transmitted],n1,&0)`;;

let VALID_OPTICAL_SYSTEM_TAC x =
  SIMP_TAC[x;is_valid_optical_system;is_valid_optical_component;ALL;
    is_valid_free_space;is_valid_interface]
  THEN CONV_TAC REAL_FIELD;;


let VALID_THIN_LENS = prove
  (`!R1 R2 n1 n2.
    ~(R1= &0) /\ ~(R2= &0) /\ &0 < n1 /\ &0 < n2 ==>
      is_valid_optical_system (thin_lens R1 R2 n1 n2)`,
  VALID_OPTICAL_SYSTEM_TAC thin_lens);;


let THIN_LENS_MAT = prove
  (`!R1 R2 n1 n2.
    ~(R1 = &0) /\ ~(R2 = &0) /\ &0 < n1 /\ &0 < n2 ==>
      system_composition (thin_lens R1 R2 n1 n2) = 
      cmat2x2 (Cx(&1)) (Cx(&0)) (Cx(((n2 - n1)/n1) *(&1/R2 - &1/R1))) (Cx(&1))`,

REWRITE_TAC[system_composition;head_index;interface_matrix;free_space_matrix;
            thin_lens;CMAT2X2_MUL;CMAT2X2_EQ;COMPLEX_MUL_LID;COMPLEX_MUL_RID;
            COMPLEX_MUL_LZERO;COMPLEX_MUL_RZERO] THEN 
REWRITE_TAC[GSYM CX_ADD;GSYM CX_MUL] THEN 
REWRITE_TAC[CX_INJ]
THEN CONV_TAC REAL_FIELD);;


let QUASI_TRANSFORMATION_LENS = 

ISPECL [`thin_lens R1 R2 n1 n2: optical_system`; 
        `gbeam:gbeam`;
        `L1:real`;
        `L2:real`;
        `lam:real`; 
        `n1:real`;
        `n2:real`;
        `w0_in:real`;
        `w0_out:real`;
        `z:real`;
        `z_n:real`; 
        `&1:real`;
        `&0:real`; 
        `(((n2 - n1)/n1) *(&1/R2 - &1/R1)):real`;
        `&1:real`] 
QUASI_OPTICAL_SYSTEM_BEAM_TRANSFORMATION_WOUT;;

let SHeFI_constraints = define  `
    SHeFI_constraints gbeam L1 L2 lam n1 n2 R1 R2 =
    ((sys_constraints
         (quasi_optical_system (thin_lens R1 R2 n1 n2) L1 L2 n1 n2)
         gbeam lam) /\
    (FST (beam_origin gbeam) = &0) /\ 
   ~(SND (beam_origin gbeam) = &0) /\
    (FST (beam_end gbeam) = &0)) `;;
 


let APEX_BEAM_WIDTH = prove (`
    !gbeam L1 L2 lam n1 n2 w0_in w0_out z z_n R1 R2.
    sys_constraints (quasi_optical_system (thin_lens R1 R2 n1 n2) L1 L2 n1 n2)
    gbeam lam /\ 
    (z = &0) /\ (z_n = &0) /\  
     z,w0_in = beam_origin gbeam /\
     z_n,w0_out = beam_end gbeam  /\ ~(w0_in = &0) /\
    ~(--((((n2 - n1) / n1 * (&1 / R2 - &1 / R1)) * L1 + &1) pow 2) =
     (((n2 - n1) / n1 * (&1 / R2 - &1 / R1)) * rayleigh_range w0_in lam) pow 2)
     ==> w0_out pow 2 / w0_in pow 2 = 
         let f = -- &1 / (((n2 - n1)/n1) *(&1/R2 - &1/R1)) in 
                    &1 / ((&1 - L1* &1 / f ) pow 2 + 
                          (--(&1 / f) * rayleigh_range w0_in lam) pow 2)`,

REPEAT STRIP_TAC THEN 
MP_TAC  QUASI_TRANSFORMATION_LENS THEN 
ANTS_TAC THEN 
ASM_SIMP_TAC[] THEN 
UNDISCH_TAC `sys_constraints
      (quasi_optical_system (thin_lens R1 R2 n1 n2) L1 L2 n1 n2)
      gbeam lam` THEN 
SIMP_TAC[sys_constraints;quasi_optical_system; is_valid_composed_system;
         thin_lens;ALL;is_valid_optical_system;is_valid_free_space; 
         is_valid_optical_component;is_valid_interface] THEN 
DISCH_TAC THEN 
ASM_SIMP_TAC[GSYM THIN_LENS_MAT;GSYM thin_lens] THEN
REWRITE_TAC[qq;RE_ADD;RE_CX;RE_MUL_II;IM_CX;REAL_FIELD ` !a. a + -- &0 = a`] THEN 
ASM_REWRITE_TAC[(REAL_FIELD ` !(a:real) (b:real). ~(a + b = &0) = ~(--a = b)`)] THEN 
SIMP_TAC[REAL_MUL_LID;REAL_MUL_LZERO;REAL_FIELD ` &1 - &0 = &1`] THEN 
LET_TAC THEN POP_ASSUM MP_TAC THEN 
SIMP_TAC[EQ_SYM_EQ] THEN DISCH_TAC THEN 
UNDISCH_TAC `~(w0_in = &0)` THEN REWRITE_TAC[ rayleigh_range] THEN 
SIMP_TAC[REAL_FIELD `(&1 - L1 * &1 / (-- &1 / ((n2 - n1) / n1 * (&1 / R2 - &1 / R1)))) = 
                     (&1 +  L1 * ((n2 - n1) / n1 * (&1 / R2 - &1 / R1)))`] THEN 
SIMP_TAC[REAL_FIELD `(--(&1 / (-- &1 / ((n2 - n1) / n1 * (&1 / R2 - &1 / R1)))) *(pi * w0_in pow 2) / lam) = 
                     (((n2 - n1) / n1 * (&1 / R2 - &1 / R1)) * w0_in pow 2 * pi / lam)`] THEN 
IMP_REWRITE_TAC [REAL_FIELD ` ~(b = &0) ==>  (c = a / b) = (b*c = a)`; 
                 REAL_FIELD ` ~(a = &0) ==> ~(a pow 2 = &0)`] THEN 
SIMP_TAC[REAL_FIELD ` a * &1 / b = a / b `] THEN
MESON_TAC[REAL_ADD_SYM;REAL_MUL_SYM]);;


let APEX_BEAM_MAGNIFICATION = prove (
    ` !gbeam L1 L2 lam n1 n2 R1 R2.
       SHeFI_constraints gbeam L1 L2 lam n1 n2 R1 R2 ==>
       (let f = -- &1 / ((n2 - n1) / n1 * (&1 / R2 - &1 / R1))  and 
         z,w0_in = beam_origin gbeam and 
         z_n,w0_out = beam_end gbeam in 
       ~(((&1 - L1 * &1 / f) pow 2 +
               (--(&1 / f) * rayleigh_range w0_in lam) pow 2) = &0)
       ==> w0_out pow 2 / w0_in pow 2 =
             &1 /((&1 - L1 * &1 / f) pow 2 +
                 (--(&1 / f) * rayleigh_range w0_in lam) pow 2))`,

REPEAT GEN_TAC THEN 
REPEAT LET_TAC THEN 
REWRITE_TAC[SHeFI_constraints] THEN 
ASM_SIMP_TAC[FST;SND] THEN 
REPEAT STRIP_TAC THEN 
MP_TAC (ISPECL 
       [`gbeam:gbeam`; `L1:real`; `L2:real`; `lam:real`; `n1:real`; `n2:real`;`w0_in:real`;
        `w0_out:real`; `z:real`; `z_n:real`; `R1:real`;`R2:real`] APEX_BEAM_WIDTH) THEN 
LET_TAC THEN 
POP_ASSUM MP_TAC THEN 
ASM_SIMP_TAC[EQ_SYM_EQ] THEN 
DISCH_TAC THEN 
ANTS_TAC THENL[
UNDISCH_TAC `~((&1 - L1 * &1 / f) pow 2 +
        (--(&1 / f) * rayleigh_range w0_in lam) pow 2 = &0)` THEN 
POP_ASSUM MP_TAC THEN SIMP_TAC[] THEN 
SIMP_TAC[GSYM (REAL_FIELD ` !(a:real) (b:real). ~(a + b = &0) = ~(--a = b)`)] THEN 
CONV_TAC REAL_FIELD;
ASM_SIMP_TAC[] ]);;

let APEX_BEAM_MAGNIFICATION_ALT = 
REWRITE_RULE [REAL_FIELD `!(a:real). a + (--b * c) pow 2 = a + (b*c) pow 2`]APEX_BEAM_MAGNIFICATION;;
